/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt;

// boilerplate!  sorry!
import scalara.ra.RA.{global => ra};

import scalanlp.stage._;
import scalanlp.stage.text._;
import scalanlp.stage.source._;

import edu.stanford.nlp.tmt.lda._;
import edu.stanford.nlp.tmt.stage.LDAStages._;
import edu.stanford.nlp.tmt.stage.GibbsLDAStages._;

import ra.pipes._;

/**
 * Scratch space for testing that the pieces in the scripts will compile.
 */
private[tmt] object Sample {

  //
  // PART 1: Initialization
  //
    
  def typeOf[M](value : M)(implicit m : scala.reflect.Manifest[M]) = m.toString;
  
  // define the text for the dataset
  val text = {
    CSVFile("isi.csv") ~> Column(4) ~>    // select column four from the file
    CaseFolder ~>                         // push everything to lowercase
    SimpleEnglishTokenizer ~>             // tokenize on spaces characters
    WordsAndNumbersOnlyFilter ~>          // ignore non-words and non-numbers
    TermCounter ~>                        // collect some term counts (needed below)
    TermMinimumLengthFilter(3) ~>         // take only terms with at least 2 characters
    TermMinimumDocumentCountFilter(5) ~>  // filter out terms that occur in less than 2 docs
    TermDynamicStopListFilter(30) ~>      // filter out the 30 most common terms in the corpus
    DocumentMinimumLengthFilter(5)        // take only documents with at least five terms
  };
  
  println(text.description);
  println();
  
  // print the text if interested
  for (words <- text.data.values.elements) { println(words.mkString(" ")); }
  
  // print the stoplist if you are curious about it
  for (term <- text.meta[TermStopListFilter].stops.elements) { println(term); }
  
  // define the dataset
  val dataset = LDADataset(text);
  println(dataset.description);
  
  // define the model parameters
  val numTopics = 30;
  val modelParams = LDA.ModelParams(numTopics);
     // this is equivalent to:
     //
     // val modelParams = LDA.ModelParams(numTopics,
     //                                   LDA.TermSmoothing(.01),
     //                                   LDA.TopicSmoothing(50.0 / numTopics));
  
  // define the training parameters
  val trainingParams = GibbsLDA.DefaultTrainingParams;
      // this is equivalent to:
      //
      // import GibbsLDATrainer._;
      // val trainingParams =
      //   TrainingParams(MaxIterations(1500),
      //                  SaveEvery(50, PerplexityEstimate, PerplexityEstimatePerDocument,
      //                                DocumentTopicDistributions,DocumentTopicAssignments));
      //
      // SaveEvery(...) could be replaced by SaveFinal() to write less output
      //
      // val trainingParams = TrainingParams(MaxIterations(1500), SaveFinal());
  
  // Name of the output model folder to generate
  val output = file("lda-"+dataset.signature+"-"+modelParams.signature);
  
  
  //
  // PART 2: Model Training
  //
  {
  // Trains the model: the model (and intermediate models) are written to the output folder,
  // which must be new
  TrainGibbsLDA(output, dataset, modelParams, trainingParams);
  

  // Load the last ten assignments 
  val perDocWordTopicProbability = LoadTrainingPerWordTopicDistributions(output, dataset, 10);
  
  DocumentTopicUsage(perDocWordTopicProbability) | CSVFile(output, "usage.csv");
  }
  
  //
  // PART 3: Inference
  //
  
  // load the trained model
  val model = LoadGibbsLDA(output);
  
  // infer topic distributions for each word in each document in the dataset.
  val perDocWordTopicProbability = InferPerWordTopicDistributions(model, dataset);
  
  //val perDocWordTopicProbability = InferPerWord(model, dataset);
  println(perDocWordTopicProbability.description);
  
  
  //
  // PART 4: Generate output
  //

  // define fields from the dataset we are going to slice against
  val school = CSVFile("isi.csv") ~> Column(12) ~> // select column 12
               Mapper { _.take(10).toString };     // and take only the first ten chars

  LDAQuery(perDocWordTopicProbability).topK(20) | CSVFile("isi-corpus-topk.csv");
               
  // create a slice object by binding the output of inference with the fields
  val slice = SlicedLDAQuery(perDocWordTopicProbability, school);
  
  // generate some output files
  slice.topK(20) | CSVFile("isi-topk.csv");
  slice.trackWords("gene","probability") | CSVFile("isi-trackwords.csv");
  slice.usage | CSVFile("isi-usage.csv");
  
  //
  // TODO
  //
  //  1. Continue training a model
  //  3. Model selection across # topics
  //  4. Topic entropy (by slice)
  //  5. Topic comparisons (by slice) [including heatmap]
  //
  
  // continue training a model
//  // continue training a
//  GibbsLDA.continue("model-"+dataset.signature+"-"+numTopics, dataset, numTopics, trainingParams);

//  //
//  // PART 5: Model selection
//  // 
//    
//  // lets take 80 percent of the input text as training data
//  val numTrain = (text.data.size * .8).toInt;
//  
//  // and build a training dataset from it
//  val training = LDADataset(text ~> Take(numTrain));
//  
//  // make sure to use the same term index when making the test dataset
//  val testing  = LDADataset(text ~> Drop(numTrain), training.meta[LDADataset.TermIndex]);
//  
//  // loop over various numbers of topics, training and evaluating each model
//  for (numTopics <- List(5,10,20,30)) {
//    val modelParams = LDA.ModelParams(numTopics);
//    val output = file("lda-"+training.signature+"-"+modelParams.signature);
//    val model = GibbsLDA.train(output, training, modelParams, trainingParams);
//
//    var alppi = 0.0;
//    ( for ((doc, dwt) <- testing.data.items.elements zip InferPerWord(model, testing).data.items.elements) yield {
//        var score = model.computeALPPI(new GibbsLDA.Document(doc.value, model.numTopics), dwt.value);
//        alppi += score;
//        score.toString;
//    } ) | file(output, "alppi-per-doc.txt");
//    
//    Iterator.single(alppi.toString) | file(output, "alppi.txt");
//  }
//  
    
  def main(args : Array[String]) { }
}
