/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt

import java.io.File;
import java.util.Collections;
import java.lang.{Iterable => JIterable};

import scalara.ra.RA.{global => ra};

import scalanlp.stage._;
import scalanlp.stage.text._;
import scalanlp.stage.source._;

import edu.stanford.nlp.tmt.lda._;
import edu.stanford.nlp.tmt.stage._;

/**
 * Global static access points for easier java interop.
 *
 * @author bethard
 */
object TMTSimple {
  
  /**
   * Create an LDA dataset from the texts, trimming terms and documents.
   *
   * @param texts The document texts, as Strings.
   * @param minTermLength The minimum number of characters a term must contain.
   * @param minTermDocCount The minimum number of documents a term must appear in.
   * @param nStopwords The number of top words by frequency to remove from all documents.
   * @param minDocTerms The minimum number of terms a document must contain.
   */
  def makeGibbsLDADataset(texts: JIterable[String], minTermLength: Int, minTermDocCount: Int, nStopwords: Int, minDocTerms: Int): LDA.Dataset = {
    val batch = Batch.fromIterable(new Java2ScalaIterable(texts));
    val dataset = LDAStages.LDADataset(
      Parcel(History.Origin("Java"), batch) ~>
      CaseFolder ~>
      SimpleEnglishTokenizer ~>
      WordsAndNumbersOnlyFilter ~>
      TermCounter ~>
      TermMinimumLengthFilter(minTermLength) ~>
      TermMinimumDocumentCountFilter(minTermDocCount) ~>
      TermDynamicStopListFilter(nStopwords) ~>
      DocumentMinimumLengthFilter(minDocTerms));
    new LDA.Dataset(dataset.history.toString, dataset.data.values,
                    dataset.meta[LDAMetadata.TermIndex].value);
  }
  
  /**
   * Train an LDA model using Gibbs sampling.
   *
   * @param dataset An LDA datset, e.g. from {@link makeGibbsLDADataset}.
   * @param nTopics The number of topics the model should learn.
   * @param outputDir The directory where the model should be written.
   */
  def trainGibbsLDAModel(dataset: LDA.Dataset, nTopics: Int, outputDir: File): Unit = {
    val modelParams = LDA.ModelParams(nTopics);
    val trainingParams = GibbsLDA.DefaultTrainingParams;
    GibbsLDA.train(outputDir, dataset, modelParams, trainingParams);
  }
  
  /**
   * Train an LDA model using Gibbs sampling.
   *
   * @param texts The document texts, as Strings. These will be converted into
   *              an LDA dataset using a set of default trimming parameters.
   * @param nTopics The number of topics the model should learn.
   * @param outputDir The directory where the model should be written.
   */
  def trainGibbsLDAModel(texts: JIterable[String], nTopics: Int, outputDir: File): Unit = {
    this.trainGibbsLDAModel(this.makeGibbsLDADataset(texts, 3, 5, 30, 5), nTopics, outputDir);
  }
  
  /**
   * Load a trained LDA model.
   *
   * @param dir The directory where the model was written.
   */
  def loadGibbsLDAModel(dir: File): GibbsLDAModel = {
    new GibbsLDAModel(GibbsLDA.loadInferenceModel(dir));
  }
  
  /**
   * A wrapper class around an LDA model, providing a simple String-based interface.
   *
   * @param model The LDA model to be wrapped.
   */
  class GibbsLDAModel(model: LDA with LDATermTopicIndex with GibbsLDA.InferenceModel) {
    
    /**
     * Get the number of topics.
     */
    def numTopics = model.numTopics;

    /**
     * Get the term index.
     */
    def termIndex = model.termIndex;

    /**
     * Get the term distribution of a single topic.
     */
    def getTermDistribution(topic: Int) = model.getTermDistribution(topic)

    /**
     * Infer document-level topics for the given texts.
     *
     * @param texts The documents whose topics are to be inferred.
	 * @return An array of topic probabilities for each text.
     */
    def infer(texts: JIterable[String]): JIterable[Array[Double]] = {
      this.inferHelper(texts, this.model.inferPerDocTopicDistribution(_, GibbsLDA.DefaultInferenceParams));
    }
  
    /**
     * Infer document-level topics for a single text.
     *
     * @param texts The document whose topics are to be inferred.
	 * @return An array of topic probabilities.
     */
    def infer(text: String): Array[Double] = {
      this.infer(Collections.singleton(text)).iterator.next;
    }
  
    /**
     * Infer word-level topics for the given texts.
     *
     * @param texts The documents whose word topics are to be inferred.
	 * @return An array of term topic probability arrays for each text.
     */
    def inferPerWord(texts: JIterable[String]): JIterable[Array[Array[Double]]] = {
      this.inferHelper(texts, this.model.inferPerDocWordTopicDistribution(_, GibbsLDA.DefaultInferenceParams));
    }
  
    /**
     * Infer word-level topics for a single text.
     *
     * @param texts The document whose word topics are to be inferred.
	 * @return An array of term topic probability arrays.
     */
    def inferPerWord(text: String): Array[Array[Double]] = {
      this.inferPerWord(Collections.singleton(text)).iterator.next;
    }

    private def inferHelper[T](texts: JIterable[String], infer: GibbsLDA.Document => T): JIterable[T] = {
      val batch = Batch.fromIterable(new Java2ScalaIterable(texts));
      val origin = Parcel(History.Origin("Java"), batch);
      val parcel = origin ~> CaseFolder ~> SimpleEnglishTokenizer
      val dataset = LDAStages.LDADataset(parcel, LDAMetadata.TermIndex(this.model.termIndex));
      val nTopics = this.model.numTopics;
      val nTerms = this.model.numTerms;
      val items = 
        for (Item(index, wordIDs) <- dataset.data.items) yield {
          val filteredIDs = wordIDs.filter(id => id >= 0 && id < nTerms);
          infer(new GibbsLDA.Document(filteredIDs, nTopics));
        }
      new Scala2JavaIterable(items);
    }
  }
    
  class Java2ScalaIterable[T](items: JIterable[T]) extends Iterable[T] {
    def elements = new Iterator[T] {
      val iter = items.iterator;
      def hasNext = iter.hasNext;
      def next = iter.next;
    }
  }
  
  class Scala2JavaIterable[T](items: Iterable[T]) extends JIterable[T] {
    def iterator = new java.util.Iterator[T] {
      val elements = items.elements;
      def hasNext = elements.hasNext;
      def next = elements.next;
      def remove = throw new java.lang.UnsupportedOperationException;
    }
  }
}
