/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt.lda;

import java.io.File;

import scalara.ra.{RA,Serializer};

import scalala.library.Vectors._;
import scalala.library.Implicits._;

import scalanlp.counters.ints.Int2IntCounter;
import scalanlp.util.Index;
import scalanlp.stage._;

import edu.stanford.nlp.tmt.stage._;

import LDA.Dataset;


/**
 * Static helper methods for interacting with the standard GibbsLDA 
 * implementation set, with limited support for customization.
 * 
 * @author dramage
 */
object GibbsLDA {
  /** Exception while running GibbsLDA. */
  class GibbsLDAException(msg : String) extends RuntimeException(msg);
  
  /**
   * Holds a bag of observations, topics for each observation, and the
   * number of times each topic shows up given those assignments.
   * 
   * @author dramage
   */
  @serializable
  @SerialVersionUID(3)
  class Document(val observed : Array[Int], val numTopics : Int) {
    /** Total length of this document (== observed.length == assigned.length) */
    val length = observed.length;
  
    /** Topic assignment for each observation: size bag length */
    val assigned = new Array[Int](length);
  
    /** Total number of times each topic seen: size numTopics - must be updated
     *  by caller to stay in sync with assignments */
    val topicTotals = new Array[Int](numTopics);
  
    /** True if assigned array contains real assignments, false if the doc has just been cleared. */
    var hasAssignments = false;
    
    // make sure to clear all assignments off the bat.
    clearAssignments();
 
    /** Clear all assignments */
    def clearAssignments() {
      java.util.Arrays.fill(assigned, -1);
      java.util.Arrays.fill(topicTotals, 0);
      hasAssignments = false;
    }
  
    /** Returns true or throws a RuntimeException */
    private[lda] def checkrep() {
      assert(observed.length == length, "Wrong lengths");
      assert(assigned.length == length, "Wrong lengths");
    
      val counts = new Array[Int](numTopics);
    
      for (i <- 0 until length) {
        assert(observed(i) >= 0, "weird observation at "+i);
      }
      
      if (hasAssignments) {
        for (i <- 0 until length) {
          assert(assigned(i) >= 0, "weird assignment at "+i);
          counts(assigned(i)) += 1;
        }
      }
    
      for (i <- 0 until topicTotals.length) {
        assert(counts(i) == topicTotals(i), "totals out of sync with observations");
      }
    }
  }
  
  /**
   * Data structures needed by the collapsed Gibbs sampler for learning
   * and inference as implemented in the GibbsLDA.LearningModel and
   * GibbsLDA.InferenceModel.  This class is not threadsafe.
   * 
   * @author dramage
   */
  trait Model extends LDA {
    val countTopicTerm : Array[Int2IntCounter.FastMapCounter] =
      Array.fromFunction(z => Int2IntCounter())(numTopics);
  
    /** Sum of topic smoothing. */
    lazy val topicSmoothDenom = sum(topicSmoothing);
  
    /** Sum of term smoothing. */
    lazy val termSmoothDenom  = sum(termSmoothing);

    checkrep();
  
    /** Checks the representation invariants */
    protected def checkrep() {
      // log("LDAModel checkrep ...");
      def check(condition : Boolean, msg : String) {
        if (!condition) throw new RuntimeException(msg);
      }
    
      check(numTopics > 0, "Must have positive number of topics");
      check(numTerms  > 0, "Must have positive number of terms");
    
      check(numTopics == topicSmoothing.length, "Wrong number of topics");
      check(numTerms == termSmoothing.length, "Wrong number of terms");
    
      check(countTopicTerm.length == numTopics, "Wrong number of topics");
    
      for (z <- 0 until numTopics) {
        check(!countTopicTerm(z).values.filter(_ < 0).hasNext,
              "betas "+z+" has negative entry");
      }
    }
    
    override def save(path : File)(implicit ra : RA) {
      import ra.pipes._;
      super.save(path);
    
      ra.log("LDA:  ... topic-term-distributions.txt");
      ( for (row <- this.countTopicTerm) yield {
        (for (t <- 0 until numTerms; val c = row(t); if c > 0) yield { t+":"+c }).mkString(" ")
      } ) | file(path,"topic-term-distributions.txt");
    }
  
    override def getTermDistribution(topic : Int) = {
      Array.fromFunction(
        i => (countTopicTerm(topic)(i) + termSmoothing(i)) /
             (countTopicTerm(topic).total + termSmoothDenom))(
        numTerms);
    }
    
    /**
     * Returns a strict interable (here a list) of Document instances from
     * the given dataset.
     */
    def mkDocuments(dataset : Dataset) : Iterable[Document] = {
      dataset.documentTerms.map(words => new Document(words, numTopics)).toList;
    }
    
    /**
     * Computes the log probability for the current document.  This measure treats
     * the assignment to theta and the model counts as observed.  Returns
     * sum_i P(w_i | theta*, beta*).
     */
    protected def computeSumLogPW(doc : Document, theta : Array[Double]) : Double = {
      // P(w_i | theta*, beta*)
      //  = \sum_z P(w_i,z_i | theta*, beta*)
      //  = \sum_z P(w_i | beta*, z_i) \cdot P(z_i | theta*)
    
      var sumLogPW = 0.0;
      for (term <- doc.observed) {
        var pW = 0.0;
        var topic = 0;
        while (topic < theta.length) {
          pW += theta(topic) * (countTopicTerm(topic)(term) + termSmoothing(term)) /
                       	       (countTopicTerm(topic).total + termSmoothDenom);
          topic += 1;
        }
        sumLogPW += Math.log(pW);
      }
      
      return sumLogPW;
    }
  }
  
  //
  // Inference
  //
  
  /** Parameters describing gibbs sambling behavior during inference. */
  case class InferenceParams(burnIterations : Int, numChains : Int,
                             samplesPerChain : Int, skipIterations : Int);
  
  /** Default value for InferenceParams. */
  val DefaultInferenceParams = InferenceParams(200,2,5,50);
  
  /**
   * The InferenceModel trait contains extra methods for doing doc-level and
   * word-level inference on a set of documents.
   * 
   * @author dramage
   */
  trait InferenceModel extends Model {
    protected val inferSampler = new InferSampler(this);
    
    /**
     * Returns the per-document distribution over topics for each document
     * in the dataset.
     */
    def inferPerDocTopicDistribution(dataset : Dataset, params : InferenceParams)
    : Iterable[Array[Double]] =
      mkDocuments(dataset).map(doc => inferPerDocTopicDistribution(doc, params));
    
    /**
     * Returns the per-document per-word distribution over topics for each
     * document in the dataset.
     */
    def inferPerDocWordTopicDistribution(dataset : Dataset, params : InferenceParams)
    : Iterable[Array[Array[Double]]] =
      mkDocuments(dataset).map(doc => inferPerDocWordTopicDistribution(doc, params));

    /**
     * Computes the per-document perplexity by first estimating
     * theta on the first half of the words of each document, and then
     * computing the likelihood under the model on the second half of
     * the words.
     */
    def computePerplexity(dataset : Dataset, params : InferenceParams) : Double = {
       var crossEntropy = 0.0;
       var wordCount = 0;

       if (!dataset.documentTerms.elements.hasNext) {
         throw new IllegalArgumentException(
           "Cannot compute perplexity on empty dataset - " +
           "are all the documents in your batch empty after filtering?");
       }
       
       for (doc <- mkDocuments(dataset); if doc.length >= 2) {
         val learnDoc = new Document(doc.observed.take(doc.length / 2), numTopics);
         val scoreDoc = new Document(doc.observed.drop(doc.length / 2), numTopics);
         
         val theta = inferPerDocTopicDistribution(learnDoc, params);
      
         crossEntropy += computeSumLogPW(scoreDoc, theta);
         wordCount += scoreDoc.length;
       }
       
       return Math.pow(2, -crossEntropy / (Math.log(2) * wordCount));
     }

    /**
     * Computes the average per-word log likelihood of the data after
     * inference.  This is just like perplexity except without leaving
     * out half the data during the estimation phase for theta.
     */
    def computeALPPI(dataset : Dataset, params : InferenceParams)
    : Iterable[Double] =
      mkDocuments(dataset).map(doc =>
        computeSumLogPW(doc, inferPerDocTopicDistribution(doc, params)));
    
    /** Inference on a single document using the given parameters. */
    def inferPerDocTopicDistribution(doc : Document, params : InferenceParams) : Array[Double] = {
      val theta = new Array[Double](numTopics);
    
      for (chain <- 0 until params.numChains) {
        doc.clearAssignments;
      
        for (iter <- 0 until (params.burnIterations-params.skipIterations)) {
          inferSampler.sample(doc);
        }
    
        // collect samples from chain
        for (sample <- 0 until params.samplesPerChain) {
          for (skip <- 0 until params.skipIterations) {
            inferSampler.sample(doc);
          }
      
          theta += doc.topicTotals;
        }
      }
    
      // normalized thetas
      theta /= sum(theta);
    
      return theta;
    }
      
    /**
     * Infers the probability of each word within the document according to
     * the given inference parameters.  The return value is an array (of length
     * doc.length) with a distribution over topics for each word in the document.
     */
    def inferPerDocWordTopicDistribution(doc : Document, params : InferenceParams) : Array[Array[Double]] = {
      val counts = Array.fromFunction(i => new Array[Double](numTopics))(doc.length);
    
      for (chain <- 0 until params.numChains) {
        doc.clearAssignments;
      
        for (iter <- 0 until (params.burnIterations-params.skipIterations)) {
          inferSampler.sample(doc);
        }
    
        // collect samples from chain
        for (sample <- 0 until params.samplesPerChain) {
          for (skip <- 0 until params.skipIterations) {
            inferSampler.sample(doc);
          }
      
          var i = 0;
          while (i < doc.length) {
            counts(i)(doc.assigned(i)) += 1;
            i += 1;
          }
        }
      }

      // normalize each component
      for (d <- counts) {
        d /= sum(d);
      }
    
      return counts;
    }
  }
  
  //
  // Learning model
  //
  
  /**
   * Parameters for training a GibbsLDA model.
   */
  case class TrainingParams(converged : LearningModel.ConvergenceTest,
                            save : LearningModel.SavePolicy);

  /**
   * Default parameters for training: every 50 iterations up to 1500, write everything.
   */
  val DefaultTrainingParams =
    TrainingParams(LearningModel.MaxIterations(1500),
                   LearningModel.SaveEvery(50, LearningModel.AllTrainingFields :_*));
  
  object LearningModel {
    /**
     * Returns true if the training should stop after now,
     * after tic iterations.
     */
    trait ConvergenceTest {
      def apply(model : LearningModel, tic : Int) : Boolean;
    }

    /** A convergence test that stops after the given number of iterations. */
    case class MaxIterations(val iterations : Int) extends ConvergenceTest {
      override def apply(model : LearningModel, tic : Int) = tic >= iterations;
    }

    /** Fields to be saved during training. */
    sealed trait TrainingField;
    case object LogProbabilityEstimate extends TrainingField;
    case object DocumentTopicDistributions extends TrainingField;
    case object DocumentTopicAssignments extends TrainingField;
  
    val AllTrainingFields = List(LogProbabilityEstimate, DocumentTopicDistributions, DocumentTopicAssignments);
  
    sealed trait SavePolicy;
    case class DontSave() extends SavePolicy;
    case class SaveFinal() extends SavePolicy;
    case class SaveEvery(iterations : Int, fields : TrainingField*) extends SavePolicy;
  }
  
  /**
   * Learning in the GibbsLDA model, updates both per-document topic
   * distributions and per-topic word distributions during each pass of Gibbs
   * sampling.
   * 
   * @author dramage
   */
  trait LearningModel extends Model {
    import LearningModel._;
    
    protected def MkLearnSampler() : LearnSampler =
      new LearnSampler(this);
    
    protected val learnSampler = MkLearnSampler();
    
    /** Trains the model on the given set of documents. */
    def train(dataset : Dataset, params : TrainingParams)(implicit ra : RA) : Unit =
      train(mkDocuments(dataset), params, 0);
    
    /**
     * Does learning on the set of documents, returning an estimate of
     * the total ALPPI based on the current point sample of theta.
     */
    protected def learn(documents : Iterable[Document]) = {
      for (doc <- documents) {
        learnSampler.sample(doc);
      }
    }
    
    /**
     * Returns an estimate of the sum of the per-word log probabilities based
     * on the current document's point assignment to z, along with the number
     * of terms that make up the sum.
     */
    protected def computeLogProbabilityEstimate(documents : Iterable[Document]) : (Double,Int) = {
      var sumLogPWNumer = 0.0;
      var sumLogPWDenom = 0;
      
      for (document <- documents) {
        val denom : Double =
          document.length + topicSmoothDenom;
      
        val theta : Array[Double] =
          Array.fromFunction(z => document.topicTotals(z) + topicSmoothing(z) / denom)(numTopics);
        
        sumLogPWNumer += computeSumLogPW(document, theta);
        sumLogPWDenom += document.length;
      }
      
      return (sumLogPWNumer, sumLogPWDenom);
    }
    
    /**
     * Trains the model on the given set of documents, starting at the given iteration
     * number (use 0 unless continuing training from a model on disk).
     */
    def train(documents : Iterable[Document], params : TrainingParams,
                        startIteration : Int)(implicit ra : RA) {
      import ra.pipes._;
    
      var tic = startIteration;
      var converged = false;
    
      def save(fields : Seq[TrainingField]) {
        val path = String.format("%05d",int2Integer(tic));

        // save model
        LDAModelIO.save(path, this);

        ra.log("LDA: Writing trainer status to "+path);

        if (fields.contains(LogProbabilityEstimate)) {
          ra.log("LDA: ... log-probability-estimate.txt");
          
          val (sumLogP, numTerms) = computeLogProbabilityEstimate(documents);
          
          Iterator.fromValues(sumLogP.toString, numTerms.toString) |
            file(path, "log-probability-estimate.txt");
        }

        if (fields.contains(DocumentTopicDistributions)) {
          ra.log("LDA: ... document-topic-distributions.tsv");
          ( for (doc <- documents) yield {
            doc.topicTotals.mkString("\t")
          } ) | file(path, "document-topic-distributions.tsv");
        }

        if (fields.contains(DocumentTopicAssignments)) {
          ra.log("LDA: ... document-topic-assignments.tsv");
          (for (doc <- documents) yield {
            doc.assigned.mkString("\t");
          } ) | file(path, "document-topic-assignments.tsv");
        }

        ra.log("LDA: done writing");
      }
    
      while (!converged) {
        learn(documents);
        converged = params.converged(this, tic);
      
        ra.log("LDA: iteration "+tic);
      
        if (converged && !params.save.isInstanceOf[DontSave]) {
          save(AllTrainingFields);
        } else {
          params.save match {
            case se : SaveEvery =>
              if (tic % se.iterations == 0) {
                save(se.fields);
              }
            
            case ds : DontSave => {
              // do nothing
            }
          
            case sae : SaveFinal => {
              // save only final model
            }
          }
        }
       
        tic += 1;
      }
    }
  }
  
  //
  // Samplers
  //
  
  @serializable
  @SerialVersionUID(2)
  private[tmt] abstract class Sampler(model : Model) {
    /**
     * Temporary array populated by sampling (not safe to use the same 
     * sampler in more than one thread).
     */
    val dist = new Array[Double](model.numTopics);
  
    /** How many times does the term appear in the topic. */
    def countTopicTerm(topic : Int, term : Int) : Int;
  
    /** How many times does the topic appear. */
    def countTopic(topic : Int) : Int;
    
    /** Update the count of the given term in the given topic by the given amount. */
    def updateTopicTerm(topic : Int, term : Int, delta : Int) : Unit;
    
    /** Creates and samples from a probability distribution over topics for a given term. */
    protected def sampleZ(doc : Document, term : Int) : Int = {
      val termSmooth = model.termSmoothing(term);
      var sumP = 0.0;
      var z = 0;
      
      while (z < model.numTopics) {
        // P(z | rest) is proportional to probability of topic in document
        // (for which no normalization is needed) times probability of word
        // in topic (for which normalization is needed).
        val pZ = (doc.topicTotals(z)+model.topicSmoothing(z)) *
                 (countTopicTerm(z,term)+termSmooth) /
                 (countTopic(z)+model.termSmoothDenom);
      
        dist(z) = pZ;
        sumP += pZ;
        z += 1;
      }
      
      return LDAModelMath.sampleFromIllDistribution(dist, sumP);
    }
  
    /** Resamples each topic assignment in the given document. */
    def sample(doc : Document) {
      assert(doc.topicTotals.length == model.numTopics);
    
      if (!doc.hasAssignments) {
        // special case sampler:  if no assignments have yet been made,
        // there are no counts to decrement before calling sampleZ.
        
        var i = 0;
        while (i < doc.length) {
          val term = doc.observed(i);
          
          // double-check the observed array in our first pass only
          assert(term >= 0 && term < model.numTerms, "Invalid document term");
      
          // sample new z from dist
          val newZ = sampleZ(doc, term);
          doc.assigned(i) = newZ;
          doc.topicTotals(newZ) += 1;
          updateTopicTerm(newZ,term,+1);
      
          i += 1;
        }
        
        doc.hasAssignments = true;
      } else {
        // regular sampler.  decrement counts, resample, then increment.
        
        var i = 0;
        while (i < doc.length) {
          val term = doc.observed(i);
        
          // remove old z from dist
          val oldZ = doc.assigned(i);
          doc.topicTotals(oldZ) -= 1;
          updateTopicTerm(oldZ,term,-1);
      
          // sample new z from dist
          val newZ = sampleZ(doc, term);
          doc.assigned(i) = newZ;
          doc.topicTotals(newZ) += 1;
          updateTopicTerm(newZ,term,+1);
      
          i += 1;
        }
      }
    }
  }
  
  /**
   * Does inference on a single document with no updates to the model.
   * 
   * @author dramage
   */
  private[tmt] class InferSampler(model : Model) extends Sampler(model) {
    override def countTopicTerm(topic : Int, term : Int) =
      model.countTopicTerm(topic)(term);
    
    override def countTopic(topic : Int) =
      model.countTopicTerm(topic).total;
    
    override def updateTopicTerm(topic : Int, term : Int, delta : Int) = {
      // intentionally empty - no updates are recorded during inference
    }
  }

  /**
   * A sampler that updates the model (normal gibbs sampling).
   * 
   * @author dramage
   */
  private[tmt] class LearnSampler(model : Model) extends Sampler(model) {
    override def countTopicTerm(topic : Int, term : Int) =
      model.countTopicTerm(topic)(term);
    
    override def countTopic(topic : Int) =
      model.countTopicTerm(topic).total;
    
    override def updateTopicTerm(topic : Int, term : Int, delta : Int) {
      model.countTopicTerm(topic)(term) += delta;
    }
  }
  
  
  //
  // convenience methods
  //
  
  /**
   * Trains a GibbsLDA model, writing the model output to the given output folder.
   * If the output folder already exists, this method looks to see if it is a
   * compatible model whose training can be continued.  The folder contains a 
   * compatible model if it contains an identical dataset description (dataset.txt),
   * and an identical set of model params (model-params.txt).  The training parameters
   * need not match - indeed you can add extra iterations onto a model that hasn't
   * quite converged.
   * 
   * @param output The name of the new output folder to create (in the current
   *   directory, as specified by the ra instance).
   * @param dataset The dataset to use while training.
   * @param modelParams Parameters describing the model to be trained.
   * @param trainingParams Parameters controlling convergence and verbosity of output
   *   during training.
   */
  def train(output : File, dataset : Dataset, modelParams : LDA.ModelParams,
            trainingParams : TrainingParams)(implicit ra : RA) :
    LDA with LDATermTopicIndex with GibbsLDA.LearningModel with GibbsLDA.InferenceModel = {
    
    import ra.pipes._;
    
    
    class LDABase {
      val termIndex = dataset.termIndex;
      
      val topicIndex = {
        Index(for (i <- 0 until modelParams.numTopics)
              yield String.format("Topic %0"+modelParams.numTopics.toString.length+"d", int2Integer(i))
        ).immutable;
      };

      val termSmoothing = modelParams.termSmoothing.dirichlet match {
        case LDA.FixedSymmetricDirichlet(pseudocount) =>
          Array.fromFunction(x => pseudocount)(termIndex.size);
      }
      
      val topicSmoothing = modelParams.topicSmoothing.dirichlet match {
        case LDA.FixedSymmetricDirichlet(pseudocount) =>
          Array.fromFunction(x => pseudocount)(topicIndex.size);
      }
    };

    // model to train (or to compare against when loading)
    val model = new LDABase with LDATermTopicIndex with GibbsLDA.LearningModel with GibbsLDA.InferenceModel;
    
    // prepare output folder
    if (!output.exists) {
      // create a new empty output directory
      ra.log("LDA: creating new model folder " + output);
      output.mkdir();
      
      // write the dataset descriptor
      Iterator.single(dataset.toString) | file(output, "dataset.txt");
    
      // write the model parameters
      Iterator.single(modelParams.toString) | file(output, "model-params.txt");
    
      // write the training parameters
      Iterator.single(trainingParams.toString) | file(output, "training-params.txt");
    
      model.train(dataset, trainingParams)(ra.branch(output));
      
      return model;
    } else if (file(output, "training-params.txt").exists && !output.list.filter(_.matches("\\d+")).isEmpty) {
        def die(msg : String) = throw new GibbsLDAException(
            "Cannot continue training model " + output + ": " + msg);
        
      // try to re-load existing model: must check dataset.txt and model-params.txt
      if (file(output, "dataset.txt").getLines.mkString("\n") != dataset.toString) {
        die("model on disk was trained on a different dataset");
      } else if (file(output, "model-params.txt").getLines.mkString("\n") != modelParams.toString) {
        die("model on disk was trained with different parameters");
      } else {
        val continued = output.listFiles.filter(_.getName.matches("\\d+")).
          toList.sort(_.getName < _.getName).last;
        
        // load model for continued training
        val loaded = LDAModelIO.load[LDA with LDATermTopicIndex with GibbsLDA.LearningModel
                                     with GibbsLDA.InferenceModel](continued);
        
        // the iteration number of the continuation
        val iteration = continued.getName.toInt+1;
        
        def ne[X](a : Iterable[X], b : Iterable[X]) : Boolean = {
          val iterA = a.elements; val iterB = b.elements;
          while (iterA.hasNext && iterB.hasNext) if (iterA.next != iterB.next) return true;
          return iterA.hasNext != iterB.hasNext;
        }
        
        if (!(model.termSmoothing deepEquals loaded.termSmoothing))
          die("incompatible model specification on term smoothing");
        if (!(model.topicSmoothing deepEquals loaded.topicSmoothing))
          die("incompatible model specification on topic smoothing");
        if (ne(model.termIndex, loaded.termIndex))
          die("incompatible model specification on term index");
        if (ne(model.topicIndex, loaded.topicIndex))
          die("incompatible model specification on topic index");

        if (trainingParams.converged(loaded, iteration)) {
          ra.log("LDA: Model already converged");
        } else {
          ra.log("LDA: Reloading document assignments");
          val documents = loaded.mkDocuments(dataset);
          GibbsLDATrainer.loadDocumentAssignments(continued, documents);
        
          ra.log("LDA: continuing training from model in folder " + continued);
          loaded.train(documents, trainingParams, continued.getName.toInt+1)(ra.branch(output));
        }
        
        return loaded;
      }
    } else {
      throw new GibbsLDAException("Output folder "+output+" already exists and is not " +
                                  "a compatible model whose training can be continued.");
    }
  }
  
  /** Returns all model folders in the given directory. */
  def findModels(path : File) : List[File] = {
    if (new File(path, "term-index.txt").exists) {
      List(path);
    } else if (path == null || path.listFiles == null) {
      List[File]();
    } else {
    ( for (folder <- path.listFiles;
           if folder.isDirectory && folder.getName.matches("\\d+"))
        yield folder
    ).toList.sort(_.getName < _.getName);
    }
  }
  
  /** Loads an instance of GibbsLDAInfer from the given model path. */
  def loadInferenceModel(path : File) : LDA with GibbsLDA.InferenceModel with LDATermTopicIndex = {
    val modelDirs = findModels(path);
    if (modelDirs.isEmpty) {
      throw new java.io.FileNotFoundException(path + " is not a valid model directory.");
    }
    val modelDir = modelDirs.last;
    LDAModelIO.load[LDA with GibbsLDA.InferenceModel with LDATermTopicIndex](modelDir);
  }
}


/**
 * Companion object for GibbsLDATrainer.
 */
object GibbsLDATrainer {

  /**
   * Loads the document topic assignments from the most recent model
   * into the given GibbsLDADocument instances for resuming training.
   */
  def loadDocumentAssignments(model : java.io.File, docs : Iterable[GibbsLDA.Document])(implicit ra : RA) : Unit = {
    import ra.pipes._;
    
    val f = file(model,"document-topic-assignments.tsv");
    
    if (!f.exists) {
      throw new IllegalArgumentException("Could not find expected file "+f);
    }
    
    val docIter = docs.elements;
    val fileIter = f.getLines;
    
    var lineNo = 1;
    
    while (docIter.hasNext && fileIter.hasNext) {
      val doc = docIter.next;
      val line = fileIter.next;
      val fields = line.split("\t");
      
      java.util.Arrays.fill(doc.topicTotals, 0);
      
      if (doc.length != fields.length) {
        throw new IllegalArgumentException("Different number of words in doc on line "+lineNo+" than loaded from "+f);
      }
      
      val loaded = fields.map(_.toInt).toArray;
      for (i <- 0 until doc.length) {
        doc.assigned(i) = loaded(i);
        doc.topicTotals(loaded(i)) += 1;
      }
      
      lineNo += 1;
    }
    
    if (docIter.hasNext) {
      throw new IllegalArgumentException("More docs provided than could be loaded from file");
    }
  }
  
  def loadDocumentTopicProbabilities(path : File)(implicit ra : RA) : scala.collection.Map[String,Array[Double]] = {
    import ra.pipes._;
    
    // all model folders
    val models = GibbsLDA.findModels(path);
    
    // the number of topics
    val numTopics = file(models(0),"topic-index.txt").getLines.map(x => 1).reduceLeft(_+_);
    
    // load all docs after 500
    val results = scala.collection.mutable.HashMap[String,Array[Double]]();
    
    var numModels = 0;
    
    for (model <- models; if (model.getName.toInt > 500)) {
      numModels += 1;
      val termSmoothing = file(model, "topic-smoothing.tsv").getLines.map(_.toDouble).toList.toArray;
      for (line <- file(model, "document-topic-distributions.tsv").getLines) {
        val fields = line.split("\\t");
        val id = fields(0);
        val thisDist = fields.drop(1).map(_.toDouble).toArray;
        thisDist += (termSmoothing : scalala.tensor.dense.DenseVector);
        thisDist /= norm(thisDist,1);
        
        if (!results.contains(id)) {
          results(id) = thisDist;
        } else {
          results(id) += (thisDist : scalala.tensor.dense.DenseVector);
        }
      }
    }
    
    // divide through by number of models to get sampled average
    for ((id,dist) <- results) {
      dist /= numModels;
    }

    // return an immutable map view of the results
    return results;
  }
}
