/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt.lda;

import scalara.ra.{RA,Serializer,Signature};

import scalala.library.Vectors._;

import scalanlp.util.Index;
import scalanlp.counters.DoubleCounter;

import java.io.File;

/**
 * Base LDA object with simple query capabilities.
 *
 * @author dramage
 */
trait LDA {
  /** The number of topics in the model. */
  val numTopics : Int;

  /** The number of terms in the model. */
  val numTerms : Int;

  /** Prior counts for each term (beta in the [gibbs] model formulation). */
  val termSmoothing : Array[Double];

  /** Prior counts for each topic (alpha in the model formulation). */
  val topicSmoothing : Array[Double];

  /** Returns the distribution over terms for the given topic. */
  def getTermDistribution(topic : Int) : Array[Double];

  /** A log message. */
  protected def log(msg : String) =
    System.err.println(msg);
  
  /** Saves the model to the given path. */
  def save(path : File)(implicit ra : RA) {
    import ra.pipes._;
    
    ra.log("LDA:  ... topic-smoothing.tsv");
    this.topicSmoothing.map(_.toString) | file(path,"topic-smoothing.tsv");
      
    ra.log("LDA:  ... term-smoothing.tsv");
    this.termSmoothing.map(_.toString)  | file(path,"term-smoothing.tsv");
  }
}

/**
 * Convenience methods for loading 
 */
object LDA {
  LDAModelIO.init();

  /** Description of the Dirichlet hyperparmeters for a multinomial distribution. */
  sealed trait DirichletParameters;
  
  /** A fixed-value (symmetric) dirichlet distribution, corresponding to +psuedocount smoothing. */
  case class FixedSymmetricDirichlet(pseudocount : Double) extends DirichletParameters;
  
  sealed case class TermSmoothing(dirichlet : DirichletParameters) extends Signature;
  
  object TermSmoothing {
    def apply(pseudocount : Double) : TermSmoothing =
      TermSmoothing(FixedSymmetricDirichlet(pseudocount));
  }
  
  sealed case class TopicSmoothing(dirichlet : DirichletParameters) extends Signature;
  object TopicSmoothing {
    def apply(pseudocount : Double) : TopicSmoothing =
      TopicSmoothing(FixedSymmetricDirichlet(pseudocount));
  }
  
  /**
   * Model parameters specifying how to construct the termSmoothing and topicSmoothing
   * arrays used by the LDA model.
   */
  case class ModelParams(numTopics : Int,
                         termSmoothing : TermSmoothing,
                         topicSmoothing  : TopicSmoothing) extends Signature {
    
     override def signature =
       numTopics + "-" + termSmoothing.signature + "-" + topicSmoothing.signature;
  }
  
  /** More static constructors for ModelParams class. */
  object ModelParams {
    def apply(numTopics : Int) : ModelParams =
      ModelParams(numTopics, TermSmoothing(.01), TopicSmoothing(50.0 / numTopics));
  }
  
  
  /**
   * A dataset that can be used for training or inference.
   * See also the constructor methods in the companion object.
   * 
   * @param name Name of the dataset (returned in toString).
   * 
   * @param docs An iterable of documents represented by a sequence
   *   of terms in the corresponding index.
   * 
   * @param termIndex The terms refered to in docs.
   * 
   * @author dramage
   */
  case class Dataset(name : String,
                     documentTerms : Iterable[Array[Int]],
                     termIndex : Index[String]) {
    
    override def toString = name;
  }
  
  /** More static constructors for Dataset. */
  object Dataset {
    def apply(name : String, docs : Iterable[Seq[String]]) : Dataset = {
      val termIndex = 
        Index(for (doc <- docs.elements; term <- doc.elements) yield term).immutable;
      
      new Dataset(name, docs.projection.map(doc => {
        val all = doc.map(termIndex);
        all.filter(_ >= 0).toArray
      }), termIndex);
    }
  }
}

/**
 * An LDA model with a term index.
 */
trait LDATermIndex extends LDA {

  val termIndex : Index[String];

  override val numTerms = termIndex.size;

  def getIndexedTermDistribution(topic : Int) : DoubleCounter[String] = {
    val cc = DoubleCounter[String]();
    val beta = getTermDistribution(topic);
    var i = 0;
    while (i < beta.length) {
      if (beta(i) != 0) {
        cc(termIndex.get(i)) = beta(i);
      }
      i += 1;
    }
    cc;
  }
  
  /** Also saves the term index. */
  override def save(path : File)(implicit ra : RA) {
    import ra.pipes._;
    super.save(path);
    
    ra.log("LDA:  ... term-index.txt");
    ( for (term <- this.termIndex) yield term) | new File(path, "term-index.txt");
  }
}

/**
 * An LDA model with a topic index.
 */
trait LDATopicIndex extends LDA {
  val topicIndex : Index[String];
  
  override val numTopics = topicIndex.size;
  
  def getTermDistribution(topic : String) : Array[Double] =
    getTermDistribution(topicIndex.immutable.indexOf(topic));
  
  /** Also saves the topic index. */
  override def save(path : File)(implicit ra : RA) {
    import ra.pipes._;
    super.save(path);
    
    ra.log("LDA:  ... topic-index.txt");
    ( for (topic <- this.topicIndex) yield topic) | new File(path, "topic-index.txt");
  }
}

/**
 * An LDA Model with both a term and topic index.
 */
trait LDATermTopicIndex extends LDATermIndex with LDATopicIndex {
  def getIndexedTermDistribution(topic : String) : DoubleCounter[String] =
    getIndexedTermDistribution(topicIndex.immutable.indexOf(topic));
}

/**
 * LDA Model utilities.
 * 
 * @author dramage
 */
object LDAModelMath {
  val _random = RA.global.random;
  
  /**
   * Returns a sample from the given ill-conditioned probability distribution
   * (doesn't sum to 1) whose sum has been precomputed.
   */
  def sampleFromIllDistribution(values : Array[Double], sum : Double) : Int = {
    var mass = sum;
    val targetProbability = _random.nextDouble() * sum;
    var i = values.length-1;
    while (i > 0) { // ok to fall through to 0
      mass -= values(i);
      if (targetProbability >= mass) {
        return i;
      }
      i -= 1;
    }
    
    // fell through to 0
    return 0;
  }
}

/**
 * IO utilities for loading and saving LDA models to disk based
 * on a set of capabilities.
 * 
 * @author dramage
 */
object LDAModelIO {
  import java.io.File;
  
  class LoadedLDABase(val path : File)(implicit val ra : RA) {
    import ra.pipes._;
    
    val termSmoothing = {
      new File(path,"term-smoothing.tsv").getLines.map(_.toDouble).toList.toArray;
    }
    
    val topicSmoothing = {
      new File(path,"topic-smoothing.tsv").getLines.map(_.toDouble).toList.toArray;
    }
  }

  /** LDA backed by a path on disk, with log msgs sent to an RA instance. */
  trait LoadedLDA extends LDA {
    /** RA instance. */
    val ra : RA;
    
    /** Path to load from. */
    val path : File;
    
    override def log(m : String) = ra.log(m);
  }
  
  /**
   * Base trait for loading termIndex and topicIndex to support
   * an LDATermTopicIndex.
   */
  trait LoadedLDATermTopicIndexBase extends LoadedLDA {
    import ra.pipes._;
    
    val termIndex  = Index(new File(path, "term-index.txt").getLines).immutable;
    val topicIndex = Index(new File(path, "topic-index.txt").getLines).immutable;
  }
  
  /**
   * Correct mix-in order for LoadedLDATermTopicIndexBase before LDATermTopicIndex.
   */
  trait LoadedLDATermTopicIndex extends LoadedLDATermTopicIndexBase with LDATermTopicIndex {
    // intentionally empty
  }

  /**
   * Correct mix-in for LoadedGibbsLDABase before GibbsLDA.
   */
  trait LoadedGibbsLDA extends LoadedLDA with GibbsLDA.Model {
    import ra.pipes._;
    
    for ((line,topic) <- new File(path,"topic-term-distributions.txt").getLines.zipWithIndex) {
      if (line.size > 0) {      
        for (termcount <- line.split(" ")) {
          val offset = termcount.indexOf(":");
          val term = termcount.substring(0,offset).toInt;
          val count = termcount.substring(offset+1).toInt;
          countTopicTerm(topic)(term) = count;
        }
      }
    }
    
    checkrep();
  }
  
  /**
   * Simplifies a manifest by returning the minimum set of Classes (traits) that
   * cover the mixin type without repetition (i.e. by removing traits that are
   * supertypes of other traits.
   */
  private def simplify[M](manifest : scala.reflect.Manifest[M]) : Set[Class[_]] = {
    // all mixings specified in the manifest
    val mixins = List() ++ (
      for (mixin <- manifest.toString.split(" with ")) yield Class.forName(mixin)
    );
    
    // the unique set of mixins not covered as a superclass
    Set() ++ (
      for (mixin <- mixins; if (mixins - mixin).forall(c => !mixin.isAssignableFrom(c))) yield mixin);
  }

  /**
   * Loads an LDA model off disk with specifications as provided through
   * reflection on the parameter argument.  Example usage:
   *   LDAModelIO.load[LDA with GibbsLDAInfer with LDATermTopicIndex](path)
   */
  private object Loader extends Serializer.Loader {
    // all valid load strategies
    var strategies = Map[Set[Class[_]], (java.io.File, RA) => LDA]();
    def register(set : Set[Class[_]])(f : (java.io.File, RA) => LDA) =
      strategies += (set -> f);
    
    // inference with a term topic index
    register(Set(classOf[GibbsLDA.InferenceModel], classOf[LDATermTopicIndex])) {
      (path : java.io.File, ra : RA) => 
      val model = (new LoadedLDABase(path)
                   with LoadedLDATermTopicIndex
                   with LoadedGibbsLDA
                   with GibbsLDA.InferenceModel);
      model.asInstanceOf[LDA];
    }
    
    // learning with a term topic index
    register(Set(classOf[GibbsLDA.LearningModel], classOf[LDATermTopicIndex])) {
      (path : java.io.File, ra : RA) => 
      val model = (new LoadedLDABase(path)
                   with LoadedLDATermTopicIndex
                   with LoadedGibbsLDA
                   with GibbsLDA.LearningModel);
      model.asInstanceOf[LDA];
    }
    
    // labeled learning with a term topic index
    register(Set(classOf[GibbsLabeledLDA.LearningModel], classOf[LDATermTopicIndex])) {
      (path : java.io.File, ra : RA) => 
      val model = (new LoadedLDABase(path)
                   with LoadedLDATermTopicIndex
                   with LoadedGibbsLDA
                   with GibbsLabeledLDA.LearningModel);
      model.asInstanceOf[LDA];
    }
    
    // learning with a term topic index
    register(Set(classOf[GibbsLDA.LearningModel], classOf[GibbsLDA.InferenceModel], classOf[LDATermTopicIndex])) {
      (path : java.io.File, ra : RA) => 
      val model = (new LoadedLDABase(path)
                   with LoadedLDATermTopicIndex
                   with LoadedGibbsLDA
                   with GibbsLDA.LearningModel
                   with GibbsLDA.InferenceModel);
      model.asInstanceOf[LDA];
    }
    
    override def canLoad[V](implicit valType : scala.reflect.Manifest[V]) : Boolean =
      strategies.contains(simplify(valType));
      
    override def load[V](file : java.io.File)(implicit valType : scala.reflect.Manifest[V], ra : RA) : V = {
      if (!Loader.canLoad(valType)) {
        throw new IllegalArgumentException("No load strategy for "+valType+"\n\n"+
                                           "Available strategies:\n"+
                                           Loader.strategies.keys.mkString("\n"));
      }
      
      val simplified = simplify(valType);
      ra.log("Loading LDA model "+file+" as "+simplified.map(_.getName).mkString(" with "));
      strategies(simplified)(file,ra).asInstanceOf[V];
    }
  }
  
  private object Saver extends Serializer.Saver {
    override def canSave[V](implicit valType : scala.reflect.Manifest[V]) : Boolean = {
      valType.erasure.getInterfaces.contains(classOf[LDA]);
    }
    
    override def save[V](path : java.io.File, model : V)(implicit valType : scala.reflect.Manifest[V], ra : RA) : Unit = {
      import ra.pipes._;
      
      if (path.exists) {
        if (!path.isDirectory) {
          throw new IllegalArgumentException("File "+path+" already exists");
        }
        ra.log("WARNING: LDAModel path "+path+" already exists.  May over-write some files.");
      }
  
      if (!path.exists) {
        path.mkdir;
      }
      
      model match {
        case m : LDA =>
          ra.log("LDA: Writing model to "+path);
          m.save(path)(ra);
          ra.log("LDA: done writing");
        
        case _ => 
          throw new IllegalArgumentException("Cannot save non-LDA");
      }
    }
  }

  /** Does nothing -- just insures the static initializer has run. */
  private[lda] def init() { }
  
  // register the loaders and savers with RA
  Serializer.registerLoad(Loader);
  Serializer.registerSave(Saver);
  
  /**
   * Loads an LDA model off disk with specifications as provided through
   * reflection on the parameter argument.  Example usage:
   *   LDAModelIO.load[LDA with GibbsLDAInfer with LDATermTopicIndex](path)
   */
  def load[M<:LDA](path : File)(implicit manifest : scala.reflect.Manifest[M], ra : RA) : M =
    return Loader.load(path)(manifest, ra).asInstanceOf[M];
  
  /**
   * Saves the given LDA model to disk.
   */
  def save[M<:LDA](path : File, model : M)(implicit manifest : scala.reflect.Manifest[M], ra : RA) =
    Saver.save(path, model)(manifest, ra);
}
