/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt.lda;

import java.io.File;

import scalara.ra.{RA,Serializer,Signature};

import scalanlp.util.Index;

object LabeledLDA {
  /**
   * A LabeledLDA.Dataset extends LDA.Dataset but also includes a co-indexed
   * iterable of observed labels for each document.
   */
  case class Dataset(override val name : String,
                     override val documentTerms : Iterable[Array[Int]],
                     override val termIndex : Index[String],
                     documentLabels : Iterable[Array[Int]],
                     labelIndex : Index[String])
  extends LDA.Dataset(name, documentTerms, termIndex);
  
  /** More static constructors for Dataset. */
  object Dataset {
    def apply(name : String, words : Iterable[Seq[String]], labels : Iterable[Seq[String]]) : Dataset = {
      val termIndex = 
        Index(for (doc <- words.elements; term <- doc.elements) yield term).immutable;
      val labelIndex =
        Index(for (doc <- labels.elements; label <- doc.elements) yield label).immutable;
      
      new Dataset(name,
        words.projection.map(_.map(termIndex).filter(_ >= 0).toArray), termIndex,
        labels.projection.map(_.map(labelIndex).filter(_ >= 0).toList.removeDuplicates.sort(_<_).toArray), labelIndex);
    }
  }
  
  /**
   * Model parameters specifying how to construct the termSmoothing and topicSmoothing
   * arrays used by the LDA model.
   */
  case class ModelParams(termSmoothing : LDA.TermSmoothing,
                         topicSmoothing  : LDA.TopicSmoothing)
  extends Signature {
     override def signature =
       termSmoothing.signature + "-" + topicSmoothing.signature;
  }
  
  /** More static constructors for ModelParams class. */
  object ModelParams {
    def apply() : ModelParams =
      ModelParams(LDA.TermSmoothing(.01), LDA.TopicSmoothing(0.1));
  }
}

/**
 * Implementation of Labeled LDA (also known as Blatant Dirichlet Allocation)
 * as described in Ramage, Hall, Nallapati, and Manning EMNLP 2009.
 * 
 * @author dramage
 */
object GibbsLabeledLDA {
  /**
   * An document that incorporates an observed label set for the LabeledLDA
   * sampler.
   * 
   * @author dramage
   */
  class Document(
    override val observed : Array[Int],
    override val numTopics : Int,
    val activeTopics : Array[Int])
  extends GibbsLDA.Document(observed, numTopics) {
  
    if (activeTopics.length == 0 || activeTopics.length > numTopics)
      throw new IllegalArgumentException("Expected activeTopics.length between 1 and numTopics");
  
    if (activeTopics.toList != activeTopics.toList.sort(_ < _).removeDuplicates)
      throw new IllegalArgumentException("Expected sorted duplicate free list");
  }

  /**
   * This LearningModel extends the standard GibbsLDA.LearningModel but using our
   * LearnSampler (below).
   * 
   * @author dramage
   */
  trait LearningModel extends GibbsLDA.LearningModel {
    override def MkLearnSampler = new LearnSampler(this);
    
    /**
     * Returns a strict interable (here a list) of Document instances from
     * the given dataset.
     */
    override def mkDocuments(dataset : LDA.Dataset) : Iterable[GibbsLDA.Document] = {
      ( for ((words,labels) <- (dataset.asInstanceOf[LabeledLDA.Dataset].documentTerms.elements zip
                                dataset.asInstanceOf[LabeledLDA.Dataset].documentLabels.elements)) 
          yield new Document(words, numTopics, labels)
      ).toList
    }
  }
  
  /**
   * The LearnSampler extends GibbsLDA.LearnSampler but constrains the set of
   * possible assignments to any given document by reading that document's activeTopics
   * array.
   * 
   * @author dramage
   */
  class LearnSampler(model : GibbsLDA.Model) extends GibbsLDA.LearnSampler(model) {
    override def sampleZ(gdoc : GibbsLDA.Document, term : Int) : Int = {
      val doc = gdoc.asInstanceOf[Document];
      
      val termSmooth = model.termSmoothing(term);
      var sumP = 0.0;
      var zI = 0;
      
      java.util.Arrays.fill(dist, 0.0);
      
      while (zI < doc.activeTopics.length) {
        val z = doc.activeTopics(zI);
        
        // P(z | rest) is proportional to probability of topic in document
        // (for which no normalization is needed) times probability of word
        // in topic (for which normalization is needed).
        val pZ = (doc.topicTotals(z)+model.topicSmoothing(z)) *
                 (countTopicTerm(z,term)+termSmooth) /
                 (countTopic(z)+model.termSmoothDenom);
        
        dist(z) = pZ;
        sumP += pZ;
        
        zI += 1;
      }
      
      return LDAModelMath.sampleFromIllDistribution(dist, sumP);
    }
  }
  
  //
  // convenience methods
  //
  
  /**
   * Trains a GibbsLabeledLDA model, writing the model output to the given output folder.
   * If the output folder already exists, this method looks to see if it is a
   * compatible model whose training can be continued.  The folder contains a 
   * compatible model if it contains an identical dataset description (dataset.txt),
   * and an identical set of model params (model-params.txt).  The training parameters
   * need not match - indeed you can add extra iterations onto a model that hasn't
   * quite converged.
   * 
   * @param output The name of the new output folder to create (in the current
   *   directory, as specified by the ra instance).
   * @param dataset The dataset to use while training.
   * @param modelParams Parameters describing the model to be trained.
   * @param trainingParams Parameters controlling convergence and verbosity of output
   *   during training.
   */
  def train(output : File, dataset : LabeledLDA.Dataset, modelParams : LabeledLDA.ModelParams,
            trainingParams : GibbsLDA.TrainingParams)(implicit ra : RA) :
    LDA with LDATermTopicIndex with LearningModel = {
      
    import ra.pipes._;
    
    class LDABase {
      val termIndex = dataset.termIndex;
      
      val topicIndex = {
        Index(dataset.labelIndex).immutable;
      };

      val termSmoothing = modelParams.termSmoothing.dirichlet match {
        case LDA.FixedSymmetricDirichlet(pseudocount) =>
          Array.fromFunction(x => pseudocount)(termIndex.size);
      }
      
      val topicSmoothing = modelParams.topicSmoothing.dirichlet match {
        case LDA.FixedSymmetricDirichlet(pseudocount) =>
          Array.fromFunction(x => pseudocount)(topicIndex.size);
      }
    };

    // model to train (or to compare against when loading)
    val model = new LDABase with LDATermTopicIndex with GibbsLabeledLDA.LearningModel with GibbsLDA.InferenceModel;
    
    // prepare output folder
    if (!output.exists) {
      // create a new empty output directory
      ra.log("LDA: creating new model folder " + output);
      output.mkdir();
      
      // write the dataset descriptor
      Iterator.single(dataset.toString) | file(output, "dataset.txt");
    
      // write the model parameters
      Iterator.single(modelParams.toString) | file(output, "model-params.txt");
    
      // write the training parameters
      Iterator.single(trainingParams.toString) | file(output, "training-params.txt");
    
      model.train(dataset, trainingParams)(ra.branch(output));
      
      return model;
    } else if (file(output, "training-params.txt").exists && !output.list.filter(_.matches("\\d+")).isEmpty) {
        def die(msg : String) = throw new GibbsLDA.GibbsLDAException(
            "Cannot continue training model " + output + ": " + msg);
        
      // try to re-load existing model: must check dataset.txt and model-params.txt
      if (file(output, "dataset.txt").getLines.mkString("\n") != dataset.toString) {
        die("model on disk was trained on a different dataset");
      } else if (file(output, "model-params.txt").getLines.mkString("\n") != modelParams.toString) {
        die("model on disk was trained with different parameters");
      } else {
        val continued = output.listFiles.filter(_.getName.matches("\\d+")).
          toList.sort(_.getName < _.getName).last;
        
        // load model for continued training
        val loaded = LDAModelIO.load[LDA with LDATermTopicIndex with GibbsLabeledLDA.LearningModel](continued);
        
        // the iteration number of the continuation
        val iteration = continued.getName.toInt+1;
        
        def ne[X](a : Iterable[X], b : Iterable[X]) : Boolean = {
          val iterA = a.elements; val iterB = b.elements;
          while (iterA.hasNext && iterB.hasNext) if (iterA.next != iterB.next) return true;
          return iterA.hasNext != iterB.hasNext;
        }
        
        if (!(model.termSmoothing deepEquals loaded.termSmoothing))
          die("incompatible model specification on term smoothing");
        if (!(model.topicSmoothing deepEquals loaded.topicSmoothing))
          die("incompatible model specification on topic smoothing");
        if (ne(model.termIndex, loaded.termIndex))
          die("incompatible model specification on term index");
        if (ne(model.topicIndex, loaded.topicIndex))
          die("incompatible model specification on topic index");

        if (trainingParams.converged(loaded, iteration)) {
          ra.log("LDA: Model already converged");
        } else {
          ra.log("LDA: Reloading document assignments");
          val documents = loaded.mkDocuments(dataset);
          GibbsLDATrainer.loadDocumentAssignments(continued, documents);
        
          ra.log("LDA: continuing training from model in folder " + continued);
          loaded.train(documents, trainingParams, continued.getName.toInt+1)(ra.branch(output));
        }
        
        return loaded;
      }
    } else {
      throw new GibbsLDA.GibbsLDAException("Output folder "+output+" already exists and is not " +
                                           "a compatible model whose training can be continued.");
    }
  }
}
