/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt.stage;

import java.io.File;

import scala.collection.mutable.HashMap;
import scala.reflect.Manifest;

import scalala.library.Vectors._;
import scalara.ra.RA;

import scalanlp.counters.DoubleCounter;
import scalanlp.collection.immutable.HTMap;

import scalanlp.stage._;
import scalanlp.stage.source.{ColumnDataSource,CSVFile};

import scalanlp.util.Index;

import edu.stanford.nlp.tmt.lda._;

//
// This class contains the main scripting interface to the toolkit.
// The wrapper classes in this package are designed to add support
// for Batch-based data collections and slicing to the raw model
// implementations in edu.stanford.nlp.tmt.lda.
//

/**
 * Metadata fields for parcels generated in LDAStages.
 */
object LDAMetadata {
  /** Metadata: what text index was used. */
  case class Text(value : Parcel[_,Batch[Seq[String]]]);

  /** Metadata: what term index was used. */
  case class TermIndex(value : Index[String]);
  
  /** Metadata: what topic index was used. */
  case class TopicIndex(value : Index[String]);
}
  
/**
 * Data and stages for working with LDA models in general.
 * 
 * @author dramage
 */
object LDAStages {
  import LDAMetadata._;
  
  /** Converts an LDADataset parcel into an LDA.Dataset */
  implicit def iParcelToLDADataset(parcel : Parcel[_ <: TermIndex,Batch[Array[Int]]])
  : LDA.Dataset =
    LDA.Dataset(parcel.history.toString, parcel.data.values, parcel.meta[TermIndex].value);
  
  /** Constructs a new lda dataset parcel from a sequence of words. */
  def LDADataset[M]
  (parcel : Parcel[M,Batch[Seq[String]]])
  (implicit m : Manifest[M], ra : RA)
  : Parcel[M with Text with TermIndex, Batch[Array[Int]]] = {
    ra.log("[LDADataset]");
    val dataset = LDA.Dataset(parcel.history.toString, parcel.data.values);
    val newMeta = parcel.meta + Text(parcel) + TermIndex(dataset.termIndex);
    Parcel(parcel.history, newMeta, parcel.data.map(doc =>
      doc.map(dataset.termIndex).filter(_ >= 0).toArray));
  }
  
  /**
   * Constructs a new lda dataset parcel from a sequence of words and a
   * pre-existing index.
   */
  def LDADataset[M]
  (parcel : Parcel[M,Batch[Seq[String]]], terms : TermIndex)
  (implicit m : Manifest[M], ra : RA)
  : Parcel[M with Text with TermIndex, Batch[Array[Int]]] = {
    ra.log("[LDADataset]");
    val newMeta = parcel.meta + Text(parcel) + terms;
    Parcel(parcel.history, newMeta, parcel.data.map(doc =>
      doc.map(terms.value).filter(_ >= 0).toArray));
  }
  
  /** Creates a SlicedLDAQuery object that computes sums across the whole corpus. */
  def LDAQuery
  (infer : Parcel[_ <: TermIndex with TopicIndex with Text,Batch[Array[Array[Double]]]])
  (implicit ra : RA)
  : SlicedLDAQuery = {
    ra.log("[LDAQuery]");
    
    val corpus = Batch.Zip(Batch.fromIterable(for (i <- 0 until infer.data.size) yield "CORPUS"));
    val topicIndex = infer.meta[TopicIndex].value;
    val slice = SlicedLDAQuery.slice(topicIndex, infer.meta[Text].value.data, infer.data, corpus);
    
    new SlicedLDAQuery(slice, topicIndex);
  }
  

  /**
   * A class that queries a set of topic probabilities as sliced by a set
   * of fields.  Use the constructor methods in the companion object to
   * instantiate.
   * 
   * @param parcel Parcel for the per-document per-word topic array
   * 
   * @author emrosen
   * @author dramage
   */
  class SlicedLDAQuery(val slices : Seq[LDASliceInfo], val topics : Index[String]) {
    import scalanlp.stage.source.CSVFile;
    
    /** Formats a group id into a raw sequence of strings suitable for building a CSV row. */
    private def format(group : Seq[Option[String]]) : Seq[String] =
      group.map(x => if (x.isDefined) x.get else "");
    
    /**
     * Prints the top-k words per topic-per-group with their probabilities.
     * 
     * Sample output:
     *   Group key 1, Group key 2, topic, Term 1, Term 2, Term 3, ...
     *   1999,alt.atheism,topic 3,debate,religion,philosophy,...
     */
    def topK(k : Int) = ColumnDataSource {
      (
        Iterator.single((1 to slices(0).group.length).map(i => "Group key "+i) ++
                        List("Topic") ++
                        (1 to k).map(i => "Term "+i))
      ) ++ (
        for (slice <- slices.elements;
             topic <- topics.elements) yield
          format(slice.group) ++ List(topic) ++ slice.perTopicWordCounts(topic).topK(k).map(_._1)
      )
    }
    
    /**
     * Tracks the given set of word through all groups.
     * 
     * Sample output:
     *   Group key 1, Group key 2, Topic, Word, Probability in topic in group, Rank in topic in group
     *   1999,alt.atheism,topic 3,TERM,(probability),(rank)
     */
    def trackWords(words : String*) = ColumnDataSource {
      (
        Iterator.single((1 to slices(0).group.length).map(i => "Group key "+i) ++
                        List("Topic", "Word", "Probability in topic in group", "Rank in topic in group"))
      ) ++ (
        for (slice <- slices.elements;
             topic <- topics.elements;
             word <- words.elements;
             val prob = slice.perTopicWordCounts(topic)(word);
             if prob > 0.0) yield {
      
          val rank = {
            var r = 1;
            for ((k,v) <- slice.perTopicWordCounts(topic).elements; if (v > prob)) { r += 1; }
            r.toString;
          };
  
          format(slice.group) ++ List(topic, word, prob.toString, rank);
        }
      )
    }
    
    /**
     * Displays total topic usage information as broken down by the given groups.
     * 
     * Sample output:
     *   group key 1, group key 2, total documents in group, topic name, fractional documents in topic in group 
     *   1999,alt.atheism,10,topic 3,1.2
     */
    def usage = ColumnDataSource {
      (
        Iterator.single((1 to slices(0).group.length).map(i => "Group key "+i) ++
                      List("Total documents in group", "Topic", "Fractional documents in topic in group"))
      ) ++ (
        for (slice <- slices.elements;
             topic <- topics.elements) yield {
          val prob = slice.perTopicWordCounts(topic).total;
          format(slice.group) ++ List(slice.numDocuments.toString, topic, prob.toString);
        }
      );
    }
  }
  
  /**
   * Info about a slice.
   * 
   * @param id Unique combination of metadata fields identifying this slcie
   * @param size Number of documents in this slice
   * @param perTopicWordCounts Map from topic to a counter of how many
   *   fractional documents each word occurs in.
   */
  case class LDASliceInfo(
    group : Seq[Option[String]], numDocuments : Int,
    perTopicWordCounts : Map[String,DoubleCounter[String]]);
  
  /**
   * Companion object constructors for SlicedLDAQuery.
   * 
   * @author emrosen
   * @author dramage
   */
  object SlicedLDAQuery {
    /** Constructor method for slicing based on a parcel. */
    def apply
    (infer : Parcel[_ <: TermIndex with TopicIndex with Text,
     Batch[Array[Array[Double]]]], fields : Parcel[_, Batch[String]] *)
    (implicit ra : RA)
    : SlicedLDAQuery = {
      ra.log("[SlicedLDAQuery]");
      
      val topicIndex = infer.meta[TopicIndex].value;
      val sliced = slice(topicIndex, infer.meta[Text].value.data, infer.data, Batch.Zip(fields.map(_.data) :_*));
      new SlicedLDAQuery(sliced, topicIndex);
    } 
    
    /**
     * Slices perDocWordTopicProbability array using the grouping information contained in
     * the given iterable of metadata.
     * 
     * @param documents document words
     * 
     * @param perDocWordTopicProbability doc -> word position -> topic -> probability.
     * 
     * @param metadata An iterable the same length as perWordThetas; for each
     *   document, records the set of annotations present on the record.
     *   Each row of metadata should be of the same fixed length in order
     *   for the output csv file to be interpretable.
     * 
     * @return A map from groups (as identified by a unique set of labels A) to
     *   map from topics to a DoubleCounter of per-topic per-word probability mass.
     */
    def apply(topicIndex : Index[String],
              documents : Batch[Seq[String]],
              perDocWordTopicProbability : Batch[Array[Array[Double]]],
              fields : Batch[Seq[Option[String]]])
    (implicit ra : RA)
    : SlicedLDAQuery = {
      ra.log("[SlicedLDAQuery]");
      
      new SlicedLDAQuery(slice(topicIndex, documents, perDocWordTopicProbability, fields), topicIndex);
    }
    
    /** Needed to sort the group names */
    private implicit def orderedOption(curr : Option[String]) = {
      new Ordered[Option[String]] {
        override def compare(that : Option[String]) = {
          if (curr.isDefined && !that.isDefined) 1;
          else if (!curr.isDefined && that.isDefined) -1;
          else if (curr.isDefined && that.isDefined) curr.get compare that.get;
          else 0;
        }
      }
    }
    
    /**
     * Slices perDocWordTopicProbability array using the grouping information contained in
     * the given iterable of metadata.
     * 
     * @param documents document words
     * 
     * @param perDocWordTopicProbability doc -> word position -> topic -> probability.
     * 
     * @param metadata An iterable the same length as perWordThetas; for each
     *   document, records the set of annotations present on the record.
     *   Each row of metadata should be of the same fixed length in order
     *   for the output csv file to be interpretable.
     * 
     * @return SliceInfo object representing information about each slice.
     */
    def slice(topicIndex : Index[String],
              documents : Batch[Seq[String]],
              perDocWordTopicProbability : Batch[Array[Array[Double]]],
              fields : Batch[Seq[Option[String]]]) : Seq[LDASliceInfo] = {
      
      // Map from groups (as identified by a unique set of labels A) to a map
      // from topics to a DoubleCounter of per-topic per-word probability mass.
      val groupCounters = new scala.collection.mutable.HashMap[Seq[Option[String]], Map[String,DoubleCounter[String]]];
      val groupSize = scalanlp.counters.IntCounter[Seq[Option[String]]]();
      
      def mkGroupCounter() =
        Map() ++ (for(topic <- topicIndex) yield (topic -> (DoubleCounter[String]() : DoubleCounter[String])));
      
      var i = 0;
      
      for ((docOpt,wordTopicArrayOpt,groupOpt) <-
             Batch.zip(documents, perDocWordTopicProbability, fields).values.elements;
           if (docOpt.isDefined && groupOpt.isDefined && wordTopicArrayOpt.isDefined)) {
        
        i += 1;
        
        val doc = docOpt.get;
        val group = groupOpt.get.toList;
        val wordTopicArray = wordTopicArrayOpt.get;
        
        if (doc.length != wordTopicArray.length)
          throw new RuntimeException("Inferred per-word topic distributions do not come from "+
                                     "the given documents.");
        
        val groupCounter = groupCounters.getOrElseUpdate(group, mkGroupCounter);
        groupSize(group) += 1;
  
        val Z = doc.length.toDouble;
        
        // add up fractional counts
        for ((term, topicArray) <- (doc.elements zip wordTopicArray.elements);
             (topicP, topicInd) <- topicArray.elements.zipWithIndex) {
          groupCounter(topicIndex.get(topicInd))(term) += (topicP / Z);
        }
      }
        
      val groups = (List() ++ groupCounters.keySet).sort(_ < _);
      return groups.map(group => LDASliceInfo(group, groupSize(group), groupCounters(group))).toList;
    }
  }
  
  /**
   * Generates per-document topic-loadings as a ColumnDataSource, with one row
   * for row of the input CSV (i.e. including rows for documents that had zero
   * words after filtering).
   * 
   * Each column contains the number of fractional words assigned to each topic
   * within that document.  These numbers can be divided by their sum to get
   * the per-document distribution over topics inferred by the model for each
   * document.
   * 
   * Sample output:
   *   Topic1,Topic2,Topic3
   *   0.2,1.3,3.6,2.1
   */
  def DocumentTopicUsage
  (parcel : Parcel[_ <: TopicIndex, Batch[Array[Array[Double]]]])
  (implicit ra : RA)
  : ColumnDataSource = {
    ra.log("[DocumentTopicUsage]");
    
    val thetas : Batch[Array[Double]] =
      parcel.data.map(values => scalala.Scalala.sum(values.map(v => new scalala.tensor.dense.DenseVector(v))).toArray);
    
    return ColumnDataSource {
      Iterator.single((parcel.meta[TopicIndex].value.toList : Seq[String])) ++
      (for ((opt,i) <- thetas.options.elements.zipWithIndex) yield opt match {
        case Some(theta) => (theta.map(_.toString) : Seq[String]);
        case None => (List() : Seq[String]);
      })
    };
  }
}

/**
 * Data and stages for working with GibbsLDA in particular.
 * 
 * @author dramage
 * @author emrosen
 */
object GibbsLDAStages {
  import LDAMetadata._;
  import LDAStages._;
  
  /**
   * Trains a model on the given dataset, saving the output in the given
   * output folder using the given model and training parameters.
   */
  def TrainGibbsLDA[M](output : File,
                       dataset : Parcel[M with Text with TermIndex, Batch[Array[Int]]],
                       modelParams : LDA.ModelParams, trainingParams : GibbsLDA.TrainingParams)
  (implicit ra : RA)
  : LDA with LDATermTopicIndex with GibbsLDA.LearningModel with GibbsLDA.InferenceModel = {
    ra.log("[TrainGibbsLDA]");
    GibbsLDA.train(output, dataset, modelParams, trainingParams)(ra);
  }
  
  /**
   * Loads a model that can be used for inference from the given saved
   * model folder.
   */
  def LoadGibbsLDA(model : File)(implicit ra : RA)
  : LDA with LDATermTopicIndex with GibbsLDA.InferenceModel = {
    ra.log("[LoadGibbsLDA]");
    GibbsLDA.loadInferenceModel(model);
  }
  
  /**
   * Loads the per-document per-word topic assignments created while training
   * the given model file on the given dataset, averaging together the results
   * in the numSavedModelsToLoad models.  Throws an exception if the model
   * or dataset do not match the ones saved in the given training folder.
   */
  def LoadTrainingPerWordTopicDistributions[M](
    modelTrainingFolder : File,
    dataset : Parcel[M with Text with TermIndex, Batch[Array[Int]]],
    numSavedModelsToLoad : Int)(implicit ra : RA)
  : Parcel[M with Text with TermIndex with TopicIndex, Batch[Array[Array[Double]]]] = {
    ra.log("[LoadTrainingPerWordTopicDistributions]");
    
    import ra.pipes._;
    
    if (!file(modelTrainingFolder,"dataset.txt").exists) {
      throw new GibbsLDA.GibbsLDAException("Not a model training folder.");
    }
    
    if (file(modelTrainingFolder,"dataset.txt").getLines.next != dataset.history.toString) {
      throw new GibbsLDA.GibbsLDAException("Model was not trained using the same dataset.");
    }
    
    val models = GibbsLDA.findModels(modelTrainingFolder);
    
    if (models.length < numSavedModelsToLoad) {
      throw new GibbsLDA.GibbsLDAException("Not enough models saved during training.");
    }
    
    val perWordThetas = new Array[Array[Double]](dataset.data.size);

    val topicIndex = Index(file(models.last, "topic-index.txt").getLines).immutable;
    
    // number of topics in the loaded model
    val numTopics = topicIndex.size;
    
    // iterator of pairs of original item and corresponding strings from file
    val itemAssignmentPairs = new Iterable[(Item[Array[Int]], List[String])] {
      override def elements = new Iterator[(Item[Array[Int]], List[String])] {
        val itemIterator = dataset.data.items.elements;
        
        // iterators of file lines loaded from disk
        val documentTopicAssignmentIterators : List[Iterator[String]] = List() ++ (
          for (folder <- models.drop(models.length-numSavedModelsToLoad)) yield
            file(folder, "document-topic-assignments.tsv").getLines);
        
        // make sure all iterators are same size
        def check() {
          var gotit = itemIterator.hasNext;
          if (!documentTopicAssignmentIterators.forall(_.hasNext == gotit)) {
            throw new GibbsLDA.GibbsLDAException("Wrong number of lines in file");
          }
        }
        
        override def hasNext = {
          check();
          itemIterator.hasNext;
        }
        
        override def next = {
          check();
          (itemIterator.next, documentTopicAssignmentIterators.map(_.next));
        }
      }
    }
    
    // the amount that each loaded obervation counts for.
    val increment : Double = 1.0 / numSavedModelsToLoad;
    
    // new items based on merging existing dataset index with assignments loaded from file
    val assignments : Iterable[Item[Array[Array[Double]]]] = (
      for ((item,lines) <- itemAssignmentPairs) yield {
        val value = Array.fromFunction(i => new Array[Double](numTopics))(item.value.length);
        for (line <- lines; val fields = line.split("\\t").map(_.toInt); (z,i) <- fields.zipWithIndex) {
          if (fields.length != item.value.length) {
            throw new GibbsLDA.GibbsLDAException("Wrong number of assignments in loaded assignments file");
          }
          value(i)(z) += increment;
        }
        Item(item.number, value);
      }
    );
    
    return dataset.withMeta(TopicIndex(topicIndex)).map(words => Batch.fromItems(assignments, words.size));
  }
  
  /**
   * Does per-document per-word inference over topics that generated that word.
   * Applies to a parcel as generated by LDADataset.
   * 
   * @author dramage
   */
  def InferPerWordTopicDistributions[M <: TermIndex]
  (model : LDA with LDATermIndex with LDATopicIndex with GibbsLDA.InferenceModel,
   parcel : Parcel[M, Batch[Array[Int]]])
  (implicit ra : RA)
  : Parcel[M with TopicIndex, Batch[Array[Array[Double]]]] = {
    ra.log("[InferPerWordTopicDistributions]");
    if ((parcel.meta[TermIndex].value.elements zip model.termIndex.elements).contains((t : (String,String)) => t._1 != t._2)) {
      throw new IllegalArgumentException("Mismatched term index between model and data.");
    }
    
    def infer(observed : Array[Int]) : Array[Array[Double]] = {
      val doc = new GibbsLDA.Document(observed, model.numTopics);
      model.inferPerDocWordTopicDistribution(doc, GibbsLDA.DefaultInferenceParams);
    }
    
    return parcel.withMeta(TopicIndex(model.topicIndex)).map(_.map(infer).strict);
  }
}

