package edu.stanford.nlp.tmt.stage

import java.io.File;

import scala.collection.mutable.HashMap;
import scala.reflect.Manifest;

import scalala.library.Vectors._;
import scalara.ra.RA;

import scalanlp.counters.DoubleCounter;
import scalanlp.collection.immutable.HTMap;

import scalanlp.stage._;
import scalanlp.stage.source.{ColumnDataSource,CSVFile};

import scalanlp.util.Index;

import edu.stanford.nlp.tmt.lda._;

/**
 * Common metadata for LabeledLDAStages.
 * 
 * @author dramage
 */
object LabeledLDAMetadata {
  /** Holds the original labels on a document set*/
  case class Labels(value : Parcel[_,Batch[Seq[String]]]);
  
  /**
   * Index from label to label number.  In general, might be different
   * from the TopicIndex if there, e.g. shared topics.
   */
  case class LabelIndex(value : Index[String]);
}

/**
 * Stages for LabeledLDA.
 * 
 * @author dramage
 */
object LabeledLDAStages {
  import LDAMetadata._;
  import LabeledLDAMetadata._;
  
  /** Converts an LDADataset parcel into an LDA.Dataset */
  implicit def iParcelToLabeledLDADataset(parcel : Parcel[_ <: Text with Labels with TermIndex with LabelIndex, Batch[(Array[Int],Array[Int])]])
  : LabeledLDA.Dataset =
    LabeledLDA.Dataset(parcel.history.toString + "\n" + parcel.meta[Labels].value.history.toString,
                       parcel.data.values.map(_._1), parcel.meta[TermIndex].value,
                       parcel.data.values.map(_._2), parcel.meta[LabelIndex].value);
  
  /** Constructs a new lda dataset parcel from a sequence of words. */
  def LabeledLDADataset[M1,M2]
  (words : Parcel[M1,Batch[Seq[String]]], labels : Parcel[M2,Batch[Seq[String]]])
  (implicit m1 : Manifest[M1], m2 : Manifest[M2], ra : RA)
  : Parcel[M1 with Text with Labels with TermIndex with LabelIndex, Batch[(Array[Int],Array[Int])]] = {
    ra.log("[LabeledLDADataset]");
    val dataset = LabeledLDA.Dataset(words.history.toString, words.data.values, labels.data.values);
    val newMeta = words.meta + Text(words) + Labels(labels) + TermIndex(dataset.termIndex) + LabelIndex(dataset.labelIndex);
    val wordBatch  =  words.data.map(doc => doc.map(dataset.termIndex).filter(_ >= 0).toArray);
    val labelBatch = labels.data.map(doc => doc.map(dataset.labelIndex).filter(_ >= 0).toList.removeDuplicates.sort(_<_).toArray);
    Parcel(words.history, newMeta, (wordBatch zip labelBatch).filter(s => s._1.isDefined && s._2.isDefined).map(s => (s._1.get, s._2.get)));
  }
  
  /**
   * Constructs a new lda dataset parcel from a sequence of words and a
   * pre-existing index.
   */
  def LabeledLDADataset[M1,M2]
  (words  : Parcel[M1,Batch[Seq[String]]], termIndex : TermIndex,
   labels : Parcel[M2,Batch[Seq[String]]], labelIndex : LabelIndex)
  (implicit m1 : Manifest[M1], m2 : Manifest[M2], ra : RA)
  : Parcel[M1 with Text with Labels with TermIndex with LabelIndex, Batch[(Array[Int],Array[Int])]] = {
    ra.log("[LabeledLDADataset]");
    val newMeta = words.meta + Text(words) + Labels(labels) + termIndex + labelIndex;
    val wordBatch  =  words.data.map(doc => doc.map(termIndex.value).filter(_ >= 0).toArray);
    val labelBatch = labels.data.map(doc => doc.map(labelIndex.value).filter(_ >= 0).toList.removeDuplicates.sort(_<_).toArray);
    Parcel(words.history, newMeta, (wordBatch zip labelBatch).map(s => (s._1.get, s._2.get)));
  }
}

/**
 * Convenience methods for scripting interaction with the GibbsLabeledLDA implementation.
 * 
 * @author dramage
 */
object GibbsLabeledLDAStages {
  import LDAMetadata._;
  import LabeledLDAMetadata._;
  import LDAStages._;
  
  /**
   * Trains a model on the given dataset, saving the output in the given
   * output folder using the given model and training parameters.
   */
  def TrainGibbsLabeledLDA[M](output : File,
                       dataset : Parcel[M with Text with Labels with TermIndex with LabelIndex, Batch[(Array[Int],Array[Int])]],
                       modelParams : LabeledLDA.ModelParams, trainingParams : GibbsLDA.TrainingParams)
  (implicit ra : RA)
  : LDA with LDATermTopicIndex with GibbsLabeledLDA.LearningModel = {
    ra.log("[TrainGibbsLabeledLDA]");
    GibbsLabeledLDA.train(output, LabeledLDAStages.iParcelToLabeledLDADataset(dataset), modelParams, trainingParams)(ra);
  }
}
